import logging

import pandas as pd

from toucan_data_sdk.utils.helpers import check_params_columns_duplicate


def compute_evolution_by_frequency(
    df,
    id_cols,
    date_col,
    value_col,
    freq=1,
    method='abs',
    format='column',
    offseted_suffix='_offseted',
    evolution_col_name='evolution_computed',
    missing_date_as_zero=False,
    raise_duplicate_error=True
):
    if missing_date_as_zero:
        how = 'outer'
        fillna = 0
    else:
        how = 'left'
        fillna = None

    return __compute_evolution(
        df=df,
        id_cols=id_cols,
        value_col=value_col,
        date_col=date_col,
        freq=freq,
        method=method,
        format=format,
        offseted_suffix=offseted_suffix,
        evolution_col_name=evolution_col_name,
        how=how,
        fillna=fillna,
        raise_duplicate_error=raise_duplicate_error
    )


def compute_evolution_by_criteria(
    df,
    id_cols,
    value_col,
    compare_to,
    method='abs',
    format='column',
    offseted_suffix='_offseted',
    evolution_col_name='evolution_computed',
    raise_duplicate_error=True
):
    return __compute_evolution(**locals())


# old name
compute_evolution = compute_evolution_by_frequency


def __compute_evolution(
    df,
    id_cols,
    value_col,
    date_col=None,
    freq=1,
    compare_to=None,
    method='abs',
    format='column',
    offseted_suffix='_offseted',
    evolution_col_name='evolution_computed',
    how='left',
    fillna=None,
    raise_duplicate_error=True
):
    """
    Compute an evolution column :
        - against a period distant from a fixed frequency.
        - against a part of the df

    Unfortunately, pandas doesn't allow .change() and .pct_change() to be
    executed with a MultiIndex.

    Args:
        df (pd.DataFrame):
        id_cols (list(str)):
        value_col (str):
        date_col (str): default None
        freq (int/pd.DateOffset/pd.Serie): default 1
        compare_to (str): default None
        method (str): default ``'abs'`` can be also ``'pct'``
        format(str): default 'column' can be also 'df'
        offseted_suffix(str): default '_offseted'
        evolution_col_name(str): default 'evolution_computed'
        how(str): default 'left'
        fillna(str/int): default None
    """
    check_params_columns_duplicate(id_cols + [value_col, date_col])
    use_date_frequency = date_col is not None
    is_freq_dict = isinstance(freq, dict)
    if use_date_frequency:
        if is_freq_dict:
            freq = pd.DateOffset(**{k: int(v) for k, v in freq.items()})
            df[date_col + '_copy'] = df[date_col]
            df[date_col] = pd.to_datetime(df[date_col])
        group_cols = id_cols + [date_col]
        df_offseted = df[group_cols + [value_col]].copy()
        df_offseted[date_col] += freq
    elif compare_to is not None:
        group_cols = id_cols
        df_offseted = df.query(compare_to).copy()
        df_offseted = df_offseted[group_cols + [value_col]]

    df_offseted_deduplicated = df_offseted.drop_duplicates(subset=group_cols)

    if df_offseted_deduplicated.shape[0] != df_offseted.shape[0] and how == 'left':
        msg = ("A dataframe for which you want to compute evolutions "
               "has duplicated values against the id_cols you indicated.")
        if raise_duplicate_error:
            raise DuplicateRowsError(msg)
        else:
            logging.getLogger(__name__).warning(f"Warning: {msg}")

    df_with_offseted_values = pd.merge(
        df,
        df_offseted_deduplicated,
        how=how,
        on=group_cols,
        suffixes=['', offseted_suffix]
    ).reset_index(drop=True)

    if fillna is not None:
        df_with_offseted_values[[value_col, value_col + offseted_suffix]] = \
            df_with_offseted_values[[value_col, value_col + offseted_suffix]].fillna(fillna)

    if use_date_frequency and is_freq_dict:
        df_with_offseted_values[date_col] = df_with_offseted_values[date_col + '_copy']
        del df_with_offseted_values[date_col + '_copy']

    if method == 'abs':
        df_with_offseted_values[evolution_col_name] = (
            df_with_offseted_values[value_col] -
            df_with_offseted_values[value_col + offseted_suffix]
        )
    elif method == 'pct':
        df_offseted_value_as_float = \
            df_with_offseted_values[value_col + offseted_suffix].astype(float)

        df_with_offseted_values[evolution_col_name] = (
            (df_with_offseted_values[value_col].astype(float) -
             df_offseted_value_as_float) /
            df_offseted_value_as_float
        )
    else:
        raise ValueError("method has to be either 'abs' or 'pct'")

    if format == 'df':
        return df_with_offseted_values
    else:
        return df_with_offseted_values[evolution_col_name]


class DuplicateRowsError(Exception):
    """Raised if a dataframe has duplicate rows"""



import os
import shutil
import tempfile

import joblib
import pytest
from requests import HTTPError

from tests.utils import DF, DF2
from toucan_data_sdk.sdk import ToucanDataSdk


def gen_client(mocker):
    class Response:
        content = 10

        def raise_for_status(self):
            pass

    resp = Response()
    client = mocker.MagicMock()
    client.sdk.post.return_value = resp
    return client


@pytest.fixture(name='sdk', scope='function')
def gen_data_sdk(mocker):
    sdk = ToucanDataSdk('some_url', auth=('', ''))
    sdk.client = gen_client(mocker)
    yield sdk
    if os.path.exists(ToucanDataSdk.EXTRACTION_CACHE_PATH):
        shutil.rmtree(ToucanDataSdk.EXTRACTION_CACHE_PATH)


def gen_client_error(mocker):
    class Response:
        content = 10

        def raise_for_status(self):
            raise HTTPError()

    resp = Response()
    client = mocker.MagicMock()
    client.sdk.post.return_value = resp
    return client


@pytest.fixture(name='sdk_error', scope='function')
def gen_data_sdk_error(mocker):
    sdk = ToucanDataSdk('some_url', auth=('', ''))
    sdk.client = gen_client_error(mocker)
    yield sdk
    if os.path.exists(ToucanDataSdk.EXTRACTION_CACHE_PATH):
        shutil.rmtree(ToucanDataSdk.EXTRACTION_CACHE_PATH)


@pytest.fixture(name='tmp_dir', scope='module')
def gen_tmp_dir():
    return tempfile.gettempdir()


@pytest.fixture(name='tmp_file', scope='function')
def gen_tmp_file(tmp_dir):
    tmp_file = tempfile.NamedTemporaryFile(dir=tmp_dir)
    yield tmp_file
    tmp_file.close()


@pytest.fixture(name='tmp_file2', scope='function')
def gen_tmp_file2(tmp_dir):
    tmp_file = tempfile.NamedTemporaryFile(dir=tmp_dir)
    yield tmp_file
    tmp_file.close()


def test_dfs(sdk, mocker):
    """It should use the cache properly"""
    mock_cache_exists = mocker.patch(
        'toucan_data_sdk.sdk.ToucanDataSdk.cache_exists')
    mock_read = mocker.patch(
        'toucan_data_sdk.sdk.ToucanDataSdk.read')
    mock_write = mocker.patch(
        'toucan_data_sdk.sdk.ToucanDataSdk.write')

    # 1. Cache directory exists
    mock_cache_exists.return_value = True
    mock_read.return_value = 1
    assert sdk.dfs == 1

    # 2. Cache directory does not exist
    mock_cache_exists.return_value = False
    mock_write.return_value = 2
    sdk._dfs = None
    assert sdk.dfs == 2

    # 3. Cache directory does not exist, keep last set value
    mock_cache_exists.return_value = False
    mock_write.return_value = 3
    assert sdk.dfs == 2


def test_dfs_http_error(sdk_error):
    """It should use the cache properly"""
    with pytest.raises(HTTPError):
        _ = sdk_error.dfs


def test_read(sdk):
    with tempfile.TemporaryDirectory() as tmp_dir:
        extraction_dir = os.path.join(tmp_dir, sdk.EXTRACTION_CACHE_PATH)
        sdk.EXTRACTION_CACHE_PATH = extraction_dir
        os.makedirs(extraction_dir)

        joblib.dump(DF, os.path.join(extraction_dir, 'a'))
        joblib.dump(DF2, os.path.join(extraction_dir, 'b'))

        dfs = sdk.read()
        assert 'a' in dfs
        assert 'b' in dfs
        assert DF.equals(dfs['a'])
        assert DF2.equals(dfs['b'])


def test_write(sdk, mocker):
    mock_extract = mocker.patch('toucan_data_sdk.sdk.extract')
    mock_extract.return_value = {'a': DF, 'b': DF2}

    with tempfile.TemporaryDirectory() as tmp_dir:
        extraction_dir = os.path.join(tmp_dir, sdk.EXTRACTION_CACHE_PATH)
        sdk.EXTRACTION_CACHE_PATH = extraction_dir

        sdk.write({'a': DF, 'b': DF2})

        assert os.path.exists(extraction_dir)
        assert 'a' in os.listdir(extraction_dir)
        assert 'b' in os.listdir(extraction_dir)


def test_cache(sdk, mocker):
    mock_read = mocker.patch('toucan_data_sdk.sdk.ToucanDataSdk.read')
    mock_write = mocker.patch('toucan_data_sdk.sdk.ToucanDataSdk.write')
    mock_path_exists = mocker.patch('os.path.exists')
    mock_is_dir = mocker.patch('os.path.isdir')

    # Cache is empty -> fill it
    mock_path_exists.return_value = False
    mock_write.return_value = {'df': DF, 'df2': DF2}

    dfs = sdk.dfs
    assert sdk.client.sdk.post.call_count == 1
    assert isinstance(dfs, dict)
    assert 'df' in dfs
    assert 'df2' in dfs

    assert DF.equals(dfs['df'])
    assert DF2.equals(dfs['df2'])

    # Cache is filled, no request to the server
    mock_path_exists.return_value = True
    mock_is_dir.return_value = True
    mock_read.return_value = {'df': DF, 'df2': DF2}

    sdk.client.sdk.post.reset_mock()
    _ = sdk.dfs
    sdk.client.sdk.post.assert_not_called()

    assert isinstance(dfs, dict)
    assert 'df' in dfs
    assert 'df2' in dfs

    assert DF.equals(dfs['df'])
    assert DF2.equals(dfs['df2'])

    mocker.stopall()


def test_invalidate_cache(sdk, mocker):
    mock_write = mocker.patch('toucan_data_sdk.sdk.ToucanDataSdk.write')
    mock_write.return_value = {'df': DF, 'df2': DF2}

    # Cache is empty -> fill it
    dfs = sdk.dfs
    assert sdk.client.sdk.post.call_count == 1
    assert isinstance(dfs, dict)
    assert 'df' in dfs
    assert 'df2' in dfs

    assert DF.equals(dfs['df'])
    assert DF2.equals(dfs['df2'])

    # Invalidate cache
    sdk.invalidate_cache()
    sdk.client.sdk.post.reset_mock()

    dfs = sdk.dfs
    assert sdk.client.sdk.post.call_count == 1
    assert isinstance(dfs, dict)
    assert 'df' in dfs
    assert 'df2' in dfs

    assert DF.equals(dfs['df'])
    assert DF2.equals(dfs['df2'])


def test_backup_existing_cache(sdk, mocker):
    mock_path_exists = mocker.patch('os.path.exists')
    mock_rmtree = mocker.patch('shutil.rmtree')
    mock_rename = mocker.patch('os.rename')

    # 1. No cache, nothing to do
    mock_path_exists.return_value = False
    sdk.backup_cache()

    assert mock_path_exists.call_count == 1
    assert mock_rmtree.call_count == 0
    assert mock_rename.call_count == 0

    mock_path_exists.reset_mock()

    # 2. Cache exists and previous backup exists
    mock_path_exists.return_value = True
    sdk.backup_cache()

    assert mock_path_exists.call_count == 2
    assert mock_rmtree.call_count == 1
    assert mock_rename.call_count == 1

    mock_path_exists.reset_mock()
    mock_rmtree.reset_mock()
    mock_rename.reset_mock()

    # 3.1 Cache exists and previous backup exists, error
    mock_path_exists.return_value = True
    mock_rmtree.side_effect = IOError('rmtree failed')
    sdk.backup_cache()

    assert mock_path_exists.call_count == 2
    assert mock_rmtree.call_count == 1
    assert mock_rename.call_count == 0

    mock_path_exists.reset_mock()
    mock_rmtree.reset_mock()
    mock_rename.reset_mock()

    # 3.2 Cache exists and previous backup exists, other error
    mock_path_exists.side_effect = [True, False]
    mock_rename.side_effect = IOError('rename failed')
    sdk.backup_cache()

    assert mock_path_exists.call_count == 2
    assert mock_rmtree.call_count == 0
    assert mock_rename.call_count == 1

    mocker.stopall()



import math
import os

import pandas as pd
import pytest

from toucan_data_sdk.utils import (
    roll_up,
    two_values_melt,
    compute_evolution_by_frequency,
    compute_evolution_by_criteria,
    compute_cumsum,
    add_missing_row,
    compute_ffill_by_group,
    aggregate_for_requesters,
    clean_dataframe)
from toucan_data_sdk.utils.compute_evolution import DuplicateRowsError
from toucan_data_sdk.utils.helpers import ParamsValueError

fixtures_base_dir = 'tests/fixtures'


def test_roll():
    """
    It should return the same DataFrame as the one after a roll_up

    """
    input_df = pd.read_csv(os.path.join(fixtures_base_dir, 'roll_up_in.csv'))
    res_df = roll_up(input_df,
                     levels=['Country', 'Region', 'City'],
                     groupby_vars=['value', 'population'],
                     value_name='Location',
                     var_name='Type')
    res_df = res_df[['Location', 'Type', 'population', 'value', 'Country', 'Region', 'City']]
    expected_output = pd.read_csv(os.path.join(fixtures_base_dir, 'roll_up.csv'))
    assert res_df.equals(expected_output)


def test_two_values_melt():
    """
    It should return the same DataFrame as the expected one in csv after a two_values_melt

    """
    input_df = pd.read_csv(os.path.join(fixtures_base_dir, 'two_values_melt_in.csv'))
    res_df = two_values_melt(
        input_df,
        first_value_vars=['avg', 'total'],
        second_value_vars=['evol_avg', 'evol_total'],
        var_name='type',
        value_name='location'
    )

    expected_output = pd.read_csv(os.path.join(fixtures_base_dir, 'two_values_melt_out.csv'))
    res_df.equals(expected_output)


def test_compute_evolution():
    """
    It should compute an evolution column against a period that is distant
    from a fixed offset
    """
    id_cols = ['City', 'Country', 'Region']
    input_df = pd.read_csv(os.path.join(fixtures_base_dir, 'compute_evolution.csv'))
    evolution_col = compute_evolution_by_frequency(
        input_df,
        id_cols,
        'Year',
        'population',
        freq=1,
        method='abs',
    )
    assert input_df['evolution'].equals(evolution_col)

    evolution_pct_col = compute_evolution_by_frequency(
        input_df,
        id_cols=id_cols,
        date_col='Year',
        value_col='population',
        freq=1,
        method='pct',
    )
    assert input_df['evolution_pct'].equals(evolution_pct_col)

    input_df = pd.read_csv(os.path.join(fixtures_base_dir, 'compute_evolution.csv'))
    evolution_df = compute_evolution_by_frequency(
        input_df,
        id_cols=id_cols,
        date_col='Year',
        value_col='population',
        freq=1,
        method='pct',
        format='df'
    )
    assert input_df['populationA-1'].equals(evolution_df['population_offseted'])
    assert input_df['evolution_pct'].equals(evolution_df['evolution_computed'])
    assert (input_df.shape[1] + 2) == evolution_df.shape[1]

    evolution_df = compute_evolution_by_frequency(
        input_df,
        id_cols=id_cols,
        date_col='Year',
        value_col='population',
        freq=1,
        method='pct',
        format='df',
        offseted_suffix='_A',
        evolution_col_name='evol'
    )
    assert input_df['populationA-1'].equals(evolution_df['population_A'])
    assert input_df['evolution_pct'].equals(evolution_df['evol'])
    assert (input_df.shape[1] + 2) == evolution_df.shape[1]

    evolution_df = compute_evolution_by_frequency(
        input_df,
        id_cols=id_cols,
        date_col='Date',
        value_col='population',
        freq={
            'years': 1
        },
        method='pct',
        format='df'
    )
    assert input_df['populationA-1'].equals(evolution_df['population_offseted'])
    assert input_df['evolution_pct'].equals(evolution_df['evolution_computed'])
    assert (input_df.shape[1] + 1) == evolution_df.shape[1]
    assert evolution_df['Date'].dtype == pd.np.object

    evolution_df = compute_evolution_by_frequency(
        input_df,
        id_cols=id_cols,
        date_col='Year',
        value_col='population',
        freq=1,
        method='abs',
        format='df',
        missing_date_as_zero=True
    )

    evolution_fillna = pd.Series([2, 10, 20, 200, 20, -13, 100, -12, -220, -7, -100])
    assert evolution_df['evolution_computed'].astype(int).equals(evolution_fillna)
    assert 11 == evolution_df.shape[0]
    assert (input_df.shape[1] + 2) == evolution_df.shape[1]


def test_compute_evolution_error_params():
    """
    It should compute an evolution column against a period that is distant
    from a fixed offset
    """
    id_cols = ['City', 'Country', 'Region', 'Year']
    input_df = pd.read_csv(os.path.join(fixtures_base_dir, 'compute_evolution.csv'))

    with pytest.raises(ParamsValueError) as e_info:
        compute_evolution_by_frequency(
            input_df,
            id_cols,
            'Year',
            'population',
            freq=1,
            method='abs',
        )

    assert "Duplicate declaration of column(s) {'Year'} in the parameters" == \
           e_info.value.args[0]


def test_compute_evolution_error_method():
    """
    It should fail to compute an evolution with unexpected method parameter.
    """
    id_cols = ['City', 'Country', 'Region']
    input_df = pd.read_csv(os.path.join(fixtures_base_dir, 'compute_evolution.csv'))
    with pytest.raises(ValueError):
        compute_evolution_by_frequency(
            input_df,
            id_cols,
            'Year',
            'population',
            freq=1,
            method='unknown',
        )


def test_compute_evolution_error_duplicate():
    """
    It should compute an evolution column against a period that is distant
    from a fixed offset
    """
    id_cols = ['Country', 'Region']
    input_df = pd.read_csv(os.path.join(fixtures_base_dir, 'compute_evolution.csv'))

    with pytest.raises(DuplicateRowsError):
        compute_evolution_by_frequency(
            input_df,
            id_cols,
            'Year',
            'population'
        )

    evolution_computed = compute_evolution_by_frequency(
        input_df,
        id_cols,
        'Year',
        'population',
        raise_duplicate_error=False
    )
    assert len(evolution_computed), 7


def test_compute_evolution_sub_df():
    """
    It should compute an evolution column against an other column
    """
    input_df = pd.read_csv(os.path.join(fixtures_base_dir, 'compute_evolution.csv'))
    evolution_df = compute_evolution_by_criteria(
        input_df,
        id_cols=['Country'],
        value_col='population',
        compare_to="City =='Nantes'",
        method='abs',
        format='df'
    )

    assert 100 == evolution_df['population_offseted'][0]
    assert input_df['evolution_nantes'].equals(evolution_df['evolution_computed'])
    assert (input_df.shape[1] + 2) == evolution_df.shape[1]


def test_compute_cumsum():
    """
    It should compute cumsum
    """
    input_df = pd.read_csv(os.path.join(fixtures_base_dir, 'compute_cumsum.csv'))
    cumsum_df = compute_cumsum(
        input_df,
        id_cols=['City', 'Country', 'Region'],
        reference_cols=['Date'],
        value_cols=['population']
    )
    assert input_df['population_cumsum'].equals(cumsum_df['population'])

    cumsum_df = compute_cumsum(
        input_df,
        id_cols=['City', 'Country', 'Region'],
        reference_cols=['Date'],
        value_cols=['population'],
        cols_to_keep=['blob']
    )
    assert input_df['blob'].equals(cumsum_df['blob'])


def test_add_missing_row():
    """
    It should add missing row compare to a reference column
    """
    input_df = pd.read_csv(
        os.path.join(fixtures_base_dir, 'add_missing_row.csv'))
    new_df = add_missing_row(
        input_df,
        id_cols=['City', 'Country', 'Region'],
        reference_col='Year'
    )
    assert len(new_df) == 12

    input_df = input_df.drop(['Country', 'Region'], axis=1).drop_duplicates()
    new_df = add_missing_row(
        input_df,
        id_cols=['City'],
        reference_col='Year'
    )
    assert len(new_df) == 12


def test_add_missing_row_use_index():
    """
    It should add missing row using the index provided
    """
    input_df = pd.read_csv(os.path.join(fixtures_base_dir, 'add_missing_row.csv'))
    new_df = add_missing_row(
        input_df,
        id_cols=['City', 'Country', 'Region'],
        reference_col='Year',
        complete_index=('2009', '2010', '2011', '2012')
    )
    assert new_df.shape[0] == 16


def test_add_missing_row_between():
    """
    It should add missing row compare to a reference column that are
    between min and max value of each group
    """
    input_df = pd.read_csv(
        os.path.join(fixtures_base_dir, 'add_missing_row.csv'))
    expected = [2011]
    new_df = add_missing_row(
        input_df,
        id_cols=['City', 'Country', 'Region'],
        reference_col='Year',
        method='between'
    )
    assert len(new_df) == 10

    result = new_df.loc[new_df['City'] == 'Nantes', 'Year'].unique().tolist()
    result.sort()
    assert result == expected


def test_add_missing_row_between_and_after():
    """
    It should add missing row compare to a reference column that are
    bigger than min of each group
    """
    input_df = pd.read_csv(
        os.path.join(fixtures_base_dir, 'add_missing_row.csv'))
    expected = [2011, 2012]
    new_df = add_missing_row(
        input_df,
        id_cols=['City', 'Country', 'Region'],
        reference_col='Year',
        method='between_and_after'
    )

    assert len(new_df) == 11

    result = new_df.loc[new_df['City'] == 'Nantes', 'Year'].unique().tolist()
    result.sort()
    assert result == expected


def test_add_missing_row_between_and_before():
    """
    It should add missing row compare to a reference column that are
    smaller than max of each group
    """
    input_df = pd.read_csv(
        os.path.join(fixtures_base_dir, 'add_missing_row.csv'))

    expected = [2010, 2011]
    new_df = add_missing_row(
        input_df,
        id_cols=['City', 'Country', 'Region'],
        reference_col='Year',
        method='between_and_before'
    )

    result = new_df.loc[new_df['City'] == 'Nantes', 'Year'].unique().tolist()
    result.sort()
    assert result == expected


def test_add_missing_row_keep_cols():
    """
    It should add missing row compare to a reference column and keep an other column
    """
    input_df = pd.read_csv(
        os.path.join(fixtures_base_dir, 'add_missing_row_2.csv'))

    new_df = add_missing_row(
        input_df,
        id_cols=['group'],
        reference_col='date',
        keep_cols=['month']
    )
    mask = (new_df['group'] == 'B') & (new_df['date'] == 20161001)
    assert len(new_df.loc[mask, 'month']) == 1
    assert new_df.loc[mask, 'month'].iloc[0] == 'octobre'
    assert math.isnan(new_df.loc[mask, 'value'].iloc[0])


def test_compute_ffill_by_group():
    """
    It should compute ffill with a groupby
    """
    input_df = pd.read_csv(os.path.join(fixtures_base_dir, 'compute_ffill_by_group.csv'))
    new_df = compute_ffill_by_group(
        input_df,
        id_cols=['City', 'Country', 'Region'],
        reference_cols=['Year'],
        value_col='population',
    )

    conditions = [
        {'Year': 2011, 'City': 'Lille', 'ffilled': True},
        {'Year': 2010, 'City': 'Nantes', 'ffilled': False},
        {'Year': 2012, 'City': 'Nantes', 'ffilled': True},
        {'Year': 2010, 'City': 'Paris', 'ffilled': False},
    ]

    for cond in conditions:
        new_df_notnull_all = new_df.loc[
            (new_df['Year'] == cond['Year']) &
            (new_df['City'] == cond['City']),
            'population'
        ].notnull().all()
        assert new_df_notnull_all == cond['ffilled']


def test_compute_ffill_by_group_2():
    """
    It should compute_ffill_by_group without messing between groups
    """
    ytd_kills = pd.DataFrame([
        {'knight': 'lancelot', 'year': 500, 'quarter': 1, 'kills': 3},
        {'knight': 'lancelot', 'year': 500, 'quarter': 2},
        {'knight': 'lancelot', 'year': 500, 'quarter': 3},
        {'knight': 'lancelot', 'year': 500, 'quarter': 4, 'kills': 4},

        {'knight': 'arthur', 'year': 500, 'quarter': 1, 'kills': 10},
        {'knight': 'arthur', 'year': 500, 'quarter': 2, 'kills': 20},
        {'knight': 'arthur', 'year': 500, 'quarter': 3},
        {'knight': 'arthur', 'year': 500, 'quarter': 4, 'kills': 30},

        {'knight': 'lancelot', 'year': 501, 'quarter': 4, 'kills': 1},
        {'knight': 'lancelot', 'year': 501, 'quarter': 1, 'kills': 0},
        {'knight': 'lancelot', 'year': 501, 'quarter': 2},
        {'knight': 'lancelot', 'year': 501, 'quarter': 3},

        {'knight': 'arthur', 'year': 501, 'quarter': 4, 'kills': 4},
        {'knight': 'arthur', 'year': 501, 'quarter': 3},
        {'knight': 'arthur', 'year': 501, 'quarter': 2, 'kills': 2},
        {'knight': 'arthur', 'year': 501, 'quarter': 1, 'kills': 0},
    ])

    filled = compute_ffill_by_group(
        ytd_kills,
        id_cols=['knight', 'year'],
        reference_cols=['quarter'],
        value_col='kills',
    )

    expected = {
        'knight == "arthur" and year == 500 and quarter == 3': 20,
        'knight == "arthur" and year == 501 and quarter == 3': 2,
        'knight == "lancelot" and year == 500 and quarter == 2': 3,
        'knight == "lancelot" and year == 500 and quarter == 3': 3,
        'knight == "lancelot" and year == 501 and quarter == 2': 0,
        'knight == "lancelot" and year == 501 and quarter == 3': 0,
    }

    for query, expected_value in expected.items():
        assert filled.query(query).iloc[0].kills == expected_value


def test_aggregate_requesters():
    """
    It should aggregate for requesters without losing NaN values
    """
    df = pd.DataFrame([
        {'year': 2017, 'filter1': 'A', 'filter2': 'C', 'value': 1},
        {'year': 2017, 'filter1': 'A', 'filter2': 'D', 'value': 5},
        {'year': 2017, 'filter1': 'B', 'filter2': pd.np.nan, 'value': 2},
        {'year': 2017, 'filter1': 'B', 'filter2': 'C', 'value': 8},
    ])
    res = aggregate_for_requesters(
        df,
        ['year'],
        {'filter1': 'All 1', 'filter2': 'All 2'},
    )
    mask = (res['filter1'] == 'All 1') & (res['filter2'] == 'All 2')
    assert res[mask]['value'][0] == 16
    mask = (res['filter1'] == 'All 1') & (res['filter2'] == 'C')
    assert res[mask]['value'][0] == 9

    res = aggregate_for_requesters(
        df,
        ['year'],
        {'filter1': 'All 1', 'filter2': 'All 2'},
        'max'
    )
    mask = (res['filter1'] == 'All 1') & (res['filter2'] == 'All 2')
    assert res[mask]['value'][0] == 8
    mask = (res['filter1'] == 'All 1') & (res['filter2'] == 'C')
    assert res[mask]['value'][0] == 8


def test_clean_dataframe():
    """It should clean a dataframe"""
    df = pd.DataFrame([
        {'DATE OF BIRTH     ': 1991.0, 'name': 'Eric', 'age': 26, 'sex': 'male'},
        {'DATE OF BIRTH     ': 1980.0, 'name': 'Samya', 'age': 37, 'sex': 'female'},
        {'DATE OF BIRTH     ': 1988.0, 'name': 'Romain', 'age': 29, 'sex': 'male'},
        {'DATE OF BIRTH     ': 1988.0, 'name': 'Fred', 'age': 29, 'sex': 'male'},
        {'DATE OF BIRTH     ': 1990.0, 'name': 'Virginie', 'age': 27, 'sex': 'female'},
        {'DATE OF BIRTH     ': 1991.0, 'name': 'Pierre', 'age': 26, 'sex': 'male'},
        {'DATE OF BIRTH     ': 1999.0, 'name': 'Erwam', 'age': 18, 'sex': 'male'},
        {'DATE OF BIRTH     ': 2000.0, 'name': 'Sophie', 'age': 17, 'sex': 'female'},
        {'DATE OF BIRTH     ': 1970.0, 'name': 'Aurelie', 'age': 47, 'sex': 'female'},
        {'DATE OF BIRTH     ': 1993.0, 'name': 'Jeremie', 'age': 24, 'sex': 'male'},
        {'DATE OF BIRTH     ': 1973.0, 'name': 'Emilie', 'age': 44, 'sex': 'female'},
    ])
    df = clean_dataframe(df, threshold=5, rename_cols={'name': 'surname'})

    assert set(df.columns) == {'date-of-birth', 'surname', 'age', 'sex'}
    assert df['date-of-birth'].dtype == 'int'
    assert df['sex'].dtype == 'category'



import logging
import os
import shutil
import tempfile
import zipfile

import joblib
from toucan_client import ToucanClient


logger = logging.getLogger(__name__)


class ToucanDataSdk:
    EXTRACTION_CACHE_PATH = 'extraction_cache'
    EXTRACTION_CACHE_PATH_BK = 'extraction_cache.bk'

    def __init__(self, small_app_url, auth, stage="staging"):
        self.client = ToucanClient(small_app_url, auth=auth)
        self.client.stage = stage
        self._dfs = None

    @property
    def dfs(self):
        if self._dfs is None:
            if self.cache_exists():
                self._dfs = self.read()
                logger.info('DataFrames extracted from cache')
            else:
                resp = self.client.sdk.post()
                resp.raise_for_status()
                self._dfs = self.write(resp.content)
                logger.info('Data fetched and cached')
        return self._dfs

    def invalidate_cache(self):
        self._dfs = None
        self.backup_cache()

    def read(self):
        """
        Returns:
            dict: Dict[str, DataFrame]
        """
        logger.info(
            'Reading data from cache ({})'.format(self.EXTRACTION_CACHE_PATH))
        return {
            name: self.read_entry(name)
            for name in os.listdir(self.EXTRACTION_CACHE_PATH)
        }

    def read_entry(self, file_name):
        """
        Args:
            file_name (str):

        Returns:
            pd.DataFrame:
        """
        file_path = os.path.join(self.EXTRACTION_CACHE_PATH, file_name)
        logger.info('Reading cache entry: {}'.format(file_path))
        return joblib.load(file_path)

    def write(self, data):
        """
        Args:
            data (str | byte):

        Returns:
            dict: Dict[str, DataFrame]
        """
        if not os.path.exists(self.EXTRACTION_CACHE_PATH):
            os.makedirs(self.EXTRACTION_CACHE_PATH)

        dfs = extract(data)
        for name, df in dfs.items():
            self.write_entry(name, df)
        return dfs

    def write_entry(self, file_name, df):
        """
        Args:
            file_name (str):
            df (DataFrame):
        """
        file_path = os.path.join(self.EXTRACTION_CACHE_PATH, file_name)
        joblib.dump(df, filename=file_path)
        logger.info('Cache entry added: {}'.format(file_path))

    def cache_exists(self):
        return os.path.exists(self.EXTRACTION_CACHE_PATH) and \
            os.path.isdir(self.EXTRACTION_CACHE_PATH)

    def backup_cache(self):
        if os.path.exists(self.EXTRACTION_CACHE_PATH):
            if os.path.exists(self.EXTRACTION_CACHE_PATH_BK):
                try:
                    shutil.rmtree(self.EXTRACTION_CACHE_PATH_BK)
                except (OSError, IOError) as e:  # For Python 2.7+ compatibility
                    logger.error('failed to remove old backup: ' + str(e))
                    return

            try:
                os.rename(self.EXTRACTION_CACHE_PATH, self.EXTRACTION_CACHE_PATH_BK)
            except (OSError, IOError) as e:
                logger.error('failed to backup current cache' + str(e))


def extract_zip(tmp_file):
    """
    Returns:
        dict: Dict[str, DataFrame]
    """
    dfs = {}
    with zipfile.ZipFile(tmp_file.name, mode='r') as z_file:
        names = z_file.namelist()
        for name in names:
            content = z_file.read(name)
            with tempfile.NamedTemporaryFile() as tmp_file:
                tmp_file.write(content)
                tmp_file.flush()
                tmp_file.seek(0)
                dfs[name] = joblib.load(tmp_file.name)
    return dfs


def extract(data):
    """
    Args:
        data (str | byte):

    Returns:
        dict: Dict[str, DataFrame]

    """
    with tempfile.NamedTemporaryFile() as tmp_file:
        tmp_file.write(data)
        tmp_file.flush()
        tmp_file.seek(0)

        if is_zipfile(tmp_file):
            return extract_zip(tmp_file)
        else:
            raise DataSdkError('Unsupported file type')


def is_zipfile(tmp_file):
    """
    Args:
        tmp_file (tempfile.TemporaryFile):

    Returns:
        bool:
    """
    try:
        return zipfile.is_zipfile(tmp_file.name)
    except Exception:
        return False
    finally:
        tmp_file.seek(0)


class DataSdkError(Exception):
    pass



from copy import copy

import pandas as pd
import pytest
from numpy import inf, nan, testing, isnan

from toucan_data_sdk.utils.postprocess import (
    replace, rename, melt, top, pivot,
    pivot_by_group, argmax, fillna,
    query_df, add, subtract, multiply,
    divide, cumsum, percentage, waterfall
)


@pytest.fixture
def sample_data():
    return [
        {'ord': 1, 'category_name': 'Clap', 'category_id': 'clap', 'product_id': 'super clap',
         'date': 't1', 'played': 12},
        {'ord': 10, 'category_name': 'Clap', 'category_id': 'clap', 'product_id': 'clap clap',
         'date': 't1', 'played': 1},
        {'ord': 1, 'category_name': 'Snare', 'category_id': 'snare', 'product_id': 'tac',
         'date': 't1', 'played': 1},
        {'ord': 1, 'category_name': 'Clap', 'category_id': 'clap', 'product_id': 'super clap',
         'date': 't2', 'played': 10},
        {'ord': 1, 'category_name': 'Snare', 'category_id': 'snare', 'product_id': 'tac',
         'date': 't2', 'played': 100},
        {'ord': 1, 'category_name': 'Tom', 'category_id': 'tom', 'product_id': 'bom',
         'date': 't2', 'played': 1}
    ]


def test_replace():
    """ It should replace data in the dataframe """
    df = pd.DataFrame([
        {'year': 2016, 'month': "jan."},
        {'year': 2016, 'month': "feb."},
        {'year': 2017, 'month': "jan."},
    ])
    config = {
        'column': 'month',
        'dst_column': 'month_int',
        'to_replace': {
            "jan.": 1,
            "feb.": 2,
            "mar.": 3,
        }
    }
    expected_result = [1, 2, 1]
    df = replace(df, **config)
    assert list(df.month_int) == expected_result


def test_rename():
    """ It should return a translated data dict """
    data = pd.DataFrame([{'hello': 'world'}])
    config = {
        'values': {'world': {'fr': 'monde'}},
        'columns': {'hello': {'fr': 'bonjour'}},
        'locale': 'fr'
    }
    data_translated = [{'bonjour': 'monde'}]
    df = rename(data, **config)
    res = [{k: v for k, v in zip(df.columns, row)} for row in df.values]
    assert res == data_translated


def test_melt():
    """ It should return result for melt """
    data = pd.DataFrame([{
        'column_1': 'S45',
        'column_2': 'lalaland',
        'info_1': 10,
        'info_2': 20,
        'info_3': None
    }])

    kwargs = {
        "id": ['column_1', 'column_2'],
        "value": ['info_1', 'info_2', 'info_3']
    }

    res = melt(data, **kwargs)
    res2 = melt(data, **kwargs, dropna=True)
    assert len(res) == 3
    assert len(res2) == 2
    assert res['variable'][0] == 'info_1'
    assert res['variable'][1] == 'info_2'
    assert res['variable'][2] == 'info_3'
    assert res['value'][0] == 10
    assert res['value'][1] == 20
    assert res['value'][2] is None
    assert 'info_1' in res2['variable'].values
    assert 'info_2' in res2['variable'].values
    assert 'info_3' not in res2['variable'].values

    bad_kwargs = {
        "id": ['column_1', 'apple'],
        "value": ['info_1', 'info_2', 'info_3']
    }
    with pytest.raises(Exception) as exc_info:
        melt(data, **bad_kwargs)
    assert str(exc_info.value) == 'Invalid configuration for melt, missing ' \
                                  'key: "[\'apple\'] not in index"'

    bad_kwargs2 = {
        "id": ['column_1', 'column_2'],
        "value": ['peach', 'banana', 'info_3']
    }
    with pytest.raises(Exception) as exc_info:
        melt(data, **bad_kwargs2)
    assert str(exc_info.value) == 'Invalid configuration for melt, missing ' \
                                  'key: "[\'peach\' \'banana\'] not in index"'


def test_top():
    """ It should return result for top """
    data = pd.DataFrame([
        {'variable': 'toto', 'Category': 1, 'value': 100},
        {'variable': 'toto', 'Category': 1, 'value': 200},
        {'variable': 'toto', 'Category': 1, 'value': 300},
        {'variable': 'lala', 'Category': 1, 'value': 100},
        {'variable': 'lala', 'Category': 1, 'value': 150},
        {'variable': 'lala', 'Category': 1, 'value': 250},
        {'variable': 'lala', 'Category': 2, 'value': 350},
        {'variable': 'lala', 'Category': 2, 'value': 450}
    ])

    # ~~~ without group ~~~
    expected = [
        {'variable': 'lala', 'Category': 2, 'value': 450},
        {'variable': 'lala', 'Category': 2, 'value': 350},
        {'variable': 'toto', 'Category': 1, 'value': 300}
    ]

    kwargs = {
        "value": 'value',
        "limit": 3,
        "order": "desc"
    }
    df = top(data, **kwargs).reset_index(drop=True)
    assert pd.DataFrame(expected).equals(df)

    # ~~~ with group ~~~
    expected = [
        {'variable': 'lala', 'Category': 1, 'value': 150},
        {'variable': 'lala', 'Category': 1, 'value': 100},
        {'variable': 'lala', 'Category': 2, 'value': 450},
        {'variable': 'lala', 'Category': 2, 'value': 350},
        {'variable': 'toto', 'Category': 1, 'value': 200},
        {'variable': 'toto', 'Category': 1, 'value': 100}
    ]
    kwargs = {
        "group": ['variable', "Category"],
        "value": 'value',
        "limit": -2,
        "order": "desc"
    }
    df = top(data, **kwargs)
    wa = [{k: v for k, v in zip(df.columns, row)} for row in df.values]
    assert wa[0].keys() == expected[0].keys()
    for i in range(len(expected)):
        assert wa[i] == expected[i]

    # ~~~ bad group ~~~
    kwargs = {
        "group": ['apple'],
        "value": 'value',
        "limit": -2,
        "order": "desc"
    }
    with pytest.raises(Exception) as exc_info:
        top(data, **kwargs)
    assert str(exc_info.value) == "Invalid configuration for top, missing key: 'apple'"


def test_pivot():
    """ It should return result for pivot """

    # ~~~~ pivot on int column ~~~
    data = pd.DataFrame([
        {'variable': 'toto', 'wave': 'wave1', 'year': 2014, 'value': 300},
        {'variable': 'toto', 'wave': 'wave1', 'year': 2015, 'value': 250},
        {'variable': 'toto', 'wave': 'wave1', 'year': 2016, 'value': 450}
    ])
    kwargs = {
        'index': ['variable', 'wave'],
        'column': 'year',
        'value': 'value'
    }
    res = pivot(data, **kwargs)
    assert res[2014][0] == 300
    assert res[2015][0] == 250
    assert res[2016][0] == 450

    # ~~~ pivot on str column ~~~
    data_obj = pd.DataFrame([
        {'variable': 'toto', 'wave': 'wave1', 'year': 2014, 'value': 'value1'},
        {'variable': 'toto', 'wave': 'wave1', 'year': 2015, 'value': 'value2'},
        {'variable': 'toto', 'wave': 'wave1', 'year': 2016, 'value': 'value3'},
        {'variable': 'toto', 'wave': 'wave1', 'year': 2016, 'value': 'value4'},
        {'variable': 'toto', 'wave': 'wave1', 'year': 2014, 'value': 'value5'}
    ])
    res = pivot(data_obj, **kwargs)
    assert res[2014][0] == 'value1 value5'
    assert res[2015][0] == 'value2'
    assert res[2016][0] == 'value3 value4'

    # ~~~ missing index ~~~
    bad_kwargs = {
        'index': ['apple'],
        'column': 'year',
        'value': 'value'
    }
    with pytest.raises(Exception) as exc_info:
        pivot(data, **bad_kwargs)
    assert str(exc_info.value) == "Invalid configuration for pivot, missing key: 'apple'"


def test_pivot_by_group():
    """ It should return result for pivot """
    data = pd.DataFrame({
        'type': ['A', 'A', 'A', 'A', 'B', 'B', 'B', 'B'],
        'kpi': ['var1', 'var1_evol', 'var2', 'var2_evol',
                'var1', 'var1_evol', 'var2', 'var2_evol'],
        'montant': [5, 0.3, 6, 0.2, 7, 0.5, 8, 0.4]
    })

    # ~~~ no id_cols ~~~
    kwargs = {
        'variable': 'kpi',
        'value': 'montant',
        'new_columns': ['value', 'variation'],
        'groups': {
            'var1': ['var1', 'var1_evol'],
            'var2': ['var2', 'var2_evol']
        }
    }
    res = pivot_by_group(copy(data), **kwargs)
    assert len(res.columns) == 3
    assert 'kpi' in res.columns
    assert len(res) == 2

    assert res.loc[res['kpi'] == 'var1', 'value'].iloc[0] == 6
    assert res.loc[res['kpi'] == 'var2', 'value'].iloc[0] == 7

    # ~~~ with id_cols ~~~
    kwargs = {
        'variable': 'kpi',
        'value': 'montant',
        'new_columns': ['value', 'variation'],
        'groups': {
            'var1': ['var1', 'var1_evol'],
            'var2': ['var2', 'var2_evol']
        },
        'id_cols': ['type']
    }
    res = pivot_by_group(data, **kwargs)
    assert len(res.columns) == 4
    assert len(res) == 4

    mask = (res['kpi'] == 'var1') & (res['type'] == 'A')
    assert res.loc[mask, 'value'].iloc[0] == 5
    mask2 = (res['kpi'] == 'var2') & (res['type'] == 'B')
    assert res.loc[mask2, 'variation'].iloc[0] == 0.4


def test_argmax():
    """ It should return result for argmax """
    data = pd.DataFrame([
        {'variable': 'toto', 'wave': 'wave1', 'year': 2014, 'value': 300},
        {'variable': 'toto', 'wave': 'wave1', 'year': 2015, 'value': 250},
        {'variable': 'toto', 'wave': 'wave1', 'year': 2016, 'value': 450}])

    kwargs = {'column': 'year'}
    res = argmax(data, **kwargs)
    res = res.reset_index(drop=True)
    assert len(res) == 1
    assert res['year'][0] == 2016
    assert res['value'][0] == 450


def test_fillna():
    """ It should return result for fillna """
    data = pd.DataFrame([
        {'variable': 'toto', 'wave': 'wave1', 'year': 2014,
         'value': 300},
        {'variable': 'toto', 'wave': 'wave1', 'year': 2015,
         'value': nan},
        {'variable': 'toto', 'wave': 'wave1', 'year': 2016,
         'value': 450}
    ])
    data2 = data.copy()

    kwargs = {
        'column': 'value',
        'value': 0
    }
    res = fillna(data, **kwargs)
    assert res['value'][1] == 0

    kwargs2 = {
        'column': 'other_column',
        'value': 5
    }

    res = fillna(data2, **kwargs2)
    assert isnan(res['value'][1])
    assert 'other_column' in res.columns
    assert all(x == 5 for x in res['other_column'])


def test_query_df():
    """ It should return a filtered data dict """
    data = pd.DataFrame([{'value': 30}, {'value': 10}, {'value': 20}])
    config = {'query': 'value > 15'}
    filtered_data = pd.DataFrame([{'value': 30}, {'value': 20}], index=[0, 2])

    df = query_df(data, **config)
    assert df.equals(filtered_data)

    # ~~~ bad config ~~~
    with pytest.raises(Exception) as exc_info:
        query_df(data, **{'query': 'pomme > 10'})
    assert str(exc_info.value) == "Invalid query: name 'pomme' is not defined"


def test_math_operations():
    """ It should return result for basic math operations """
    data = pd.DataFrame([{'value1': 10, 'value2': 20},
                         {'value1': 17, 'value2': 5}])

    kwargs = {'new_column': 'result', 'column_1': 'value1', 'column_2': 'value2'}

    res = add(data, **kwargs)
    expected_col = [30, 22]
    assert res['result'].tolist() == expected_col

    res = subtract(data, **kwargs)
    expected_col = [-10, 12]
    assert res['result'].tolist() == expected_col

    res = multiply(data, **kwargs)
    expected_col = [200, 85]
    assert res['result'].tolist() == expected_col

    res = divide(data, **kwargs)
    expected_col = [.5, 3.4]
    assert res['result'].tolist() == expected_col

    # ~~~ bad column ~~~
    bad_kwargs = {'new_column': 'result', 'column_1': 'value1', 'column_2': 'apple'}

    with pytest.raises(Exception) as exc_info:
        add(data, **bad_kwargs)
    assert str(exc_info.value) == "Invalid config for sum: 'apple'"

    with pytest.raises(Exception) as exc_info:
        subtract(data, **bad_kwargs)
    assert str(exc_info.value) == "Invalid config for subtract: 'apple'"

    with pytest.raises(Exception) as exc_info:
        multiply(data, **bad_kwargs)
    assert str(exc_info.value) == "Invalid config for multiply: 'apple'"

    with pytest.raises(Exception) as exc_info:
        divide(data, **bad_kwargs)
    assert str(exc_info.value) == "Invalid config for divide: 'apple'"


def test_cumsum():
    """ It should return result for cumsum """
    data = pd.DataFrame([
        {'index_col': 'A', 'date_col': '20170901', 'value': 10},
        {'index_col': 'A', 'date_col': '20170902', 'value': 5},
        {'index_col': 'A', 'date_col': '20170909', 'value': 3},
        {'index_col': 'A', 'date_col': '20170905', 'value': 11}
    ])
    expected_col = [10, 15, 29, 26]

    kwargs = {'new_column': 'result',
              'column': 'value',
              'index': 'index_col',
              'date_column': 'date_col',
              'date_format': '%Y%m%d'}

    res = cumsum(data, **kwargs)
    assert [*res.columns] == ['date_col', 'index_col', 'value', 'result']
    assert res['result'].tolist() == expected_col

    # same column
    kwargs = {
        'new_column': 'value',
        'column': 'value',
        'index': ['index_col'],
        'date_column': 'date_col',
        'date_format': '%Y%m%d'
    }
    res = cumsum(copy(data), **kwargs)
    assert [*res.columns] == ['date_col', 'index_col', 'value']
    assert res['value'].tolist() == expected_col

    # ~~~ bad config ~~~
    bad_kwargs = {
        'new_column': 'value',
        'column': 'value',
        'index': ['index_col'],
        'date_column': 'apple',
        'date_format': '%Y%m%d'
    }
    with pytest.raises(Exception) as exc_info:
        cumsum(data, **bad_kwargs)
    assert str(exc_info.value) == "Invalid config for cumsum: 'apple'"


def test_percentage():
    """ It should add a column `number_percentage` to the dataframe """
    data = pd.DataFrame([
        {'gender': 'male', 'sport': 'bicycle', 'number': 17},
        {'gender': 'female', 'sport': 'basketball', 'number': 17},
        {'gender': 'male', 'sport': 'basketball', 'number': 3},
        {'gender': 'female', 'sport': 'football', 'number': 7},
        {'gender': 'female', 'sport': 'running', 'number': 30},
        {'gender': 'male', 'sport': 'running', 'number': 20},
        {'gender': 'male', 'sport': 'football', 'number': 21},
        {'gender': 'female', 'sport': 'bicycle', 'number': 17}
    ])

    # ~~~ without group_cols ~~~
    kwargs = {
        'new_column': 'number_percentage',
        'column': 'number'
    }
    res = percentage(data, **kwargs)
    assert res['number_percentage'].tolist()[0] == 1700/132
    assert res['number_percentage'].tolist()[2] == 300 / 132
    assert res['number_percentage'].tolist()[4] == 3000 / 132

    # ~~~ with group_cols ~~~
    kwargs = {
        'new_column': 'number_percentage',
        'column': 'number',
        'group_cols': ['sport']
    }
    res = percentage(data, **kwargs)
    expected_col = [50, 85, 15, 25, 60, 40, 75, 50]
    assert res['number_percentage'].tolist() == expected_col


def test_waterfall(sample_data):
    """ It should return value for waterfall """
    kwargs = {
        'upperGroup': {'id': 'category_id', 'label': 'category_name'},
        'insideGroup': {'id': 'product_id', 'groupsOrder': 'ord'},
        'date': 'date',
        'value': 'played',
        'start': {'label': 'Trimestre 1', 'id': 't1'},
        'end': {'label': 'Trimester 2', 'id': 't2'},
    }

    expected = [
        {'variation': nan, 'label': 'Trimestre 1', 'value': 14.0,
         'groups': nan, 'type': nan, 'order': nan},
        {'variation': -0.23076923076923078, 'label': 'Clap', 'value': -3.0,
         'groups': 'clap', 'type': 'parent', 'order': nan},
        {'variation': -0.16666666666666666, 'label': 'super clap', 'value': -2.0,
         'groups': 'clap', 'type': 'child', 'order': 1.0},
        {'variation': -1.000000, 'label': 'clap clap', 'value': -1.0,
         'groups': 'clap', 'type': 'child', 'order': 10.0},
        {'variation': 99.0, 'label': 'Snare', 'value': 99.0,
         'groups': 'snare', 'type': 'parent', 'order': nan},
        {'variation': 99.0, 'label': 'tac', 'value': 99.0,
         'groups': 'snare', 'type': 'child', 'order': 1.0},
        {'variation': inf, 'label': 'Tom', 'value': 1.0,
         'groups': 'tom', 'type': 'parent', 'order': nan},
        {'variation': inf, 'label': 'bom', 'value': 1.0,
         'groups': 'tom', 'type': 'child', 'order': 1.0},
        {'variation': nan, 'label': 'Trimester 2', 'value': 111.0,
         'groups': nan, 'type': nan, 'order': nan}
    ]
    df = pd.DataFrame(sample_data)
    df = waterfall(df, **kwargs)
    wa = [{k: v for k, v in zip(df.columns, row)} for row in df.values]
    assert wa[0].keys() == expected[0].keys()
    for i in range(len(expected)):
        testing.assert_equal(wa[i], expected[i])


def test_waterfall_upperGroup_groupsOrder(sample_data):
    for line in sample_data:
        line['category_order'] = len(line['category_name'])
        del line['ord']

    kwargs = {
        'upperGroup': {
            'id': 'category_id',
            'label': 'category_name',
            'groupsOrder': 'category_order'
        },
        'insideGroup': {
            'id': 'product_id'
        },
        'date': 'date',
        'value': 'played',
        'start': {'label': 'Trimestre 1', 'id': 't1'},
        'end': {'label': 'Trimester 2', 'id': 't2'},
    }

    expected = [
        {'variation': nan, 'label': 'Trimestre 1', 'value': 14.0, 'groups': nan, 'type': nan,
         'order': nan},
        {'variation': inf, 'label': 'Tom', 'value': 1.0, 'groups': 'tom', 'type': 'parent',
         'order': 3.0},
        {'variation': inf, 'label': 'bom', 'value': 1.0, 'groups': 'tom', 'type': 'child',
         'order': nan},
        {'variation': -0.23076923076923078, 'label': 'Clap', 'value': -3.0, 'groups': 'clap',
         'type': 'parent', 'order': 4.0},
        {'variation': -1.0, 'label': 'clap clap', 'value': -1.0, 'groups': 'clap',
         'type': 'child', 'order': nan},
        {'variation': -0.16666666666666666, 'label': 'super clap', 'value': -2.0,
         'groups': 'clap', 'type': 'child', 'order': nan},
        {'variation': 99.0, 'label': 'Snare', 'value': 99.0, 'groups': 'snare',
         'type': 'parent', 'order': 5.0},
        {'variation': 99.0, 'label': 'tac', 'value': 99.0, 'groups': 'snare', 'type': 'child',
         'order': nan},
        {'variation': nan, 'label': 'Trimester 2', 'value': 111.0, 'groups': nan, 'type': nan,
         'order': nan}
    ]

    df = pd.DataFrame(sample_data)
    df = waterfall(df, **kwargs)
    wa = [{k: v for k, v in zip(df.columns, row)} for row in df.values]
    assert wa[0].keys() == expected[0].keys()
    for i in range(len(expected)):
        testing.assert_equal(wa[i], expected[i])


def test_waterfall_no_value_start():
    kwargs = {
        'upperGroup': {'id': 'category_id', 'label': 'category_name'},
        'insideGroup': {'id': 'product_id', 'groupsOrder': 'ord'},
        'date': 'date',
        'value': 'played',
        'start': {'label': 'Trimestre 1', 'id': 't1'},
        'end': {'label': 'Trimester 2', 'id': 't2'},
    }

    data = [
        {'ord': 1, 'category_name': 'Clap', 'category_id': 'clap',
         'product_id': 'super clap', 'date': 't2', 'played': 10},
        {'ord': 1, 'category_name': 'Snare', 'category_id': 'snare',
         'product_id': 'tac', 'date': 't2', 'played': 100},
        {'ord': 1, 'category_name': 'Tom', 'category_id': 'tom',
         'product_id': 'bom', 'date': 't2', 'played': 1}
    ]

    expected = [
        {'variation': nan, 'label': 'Trimestre 1', 'value': 0, 'groups': nan, 'type': nan,
         'order': nan},
        {'variation': inf, 'label': 'Clap', 'value': 10, 'groups': 'clap', 'type': 'parent',
         'order': nan},
        {'variation': inf, 'label': 'super clap', 'value': 10, 'groups': 'clap',
         'type': 'child', 'order': 1.0},
        {'variation': inf, 'label': 'Snare', 'value': 100.0, 'groups': 'snare',
         'type': 'parent', 'order': nan},
        {'variation': inf, 'label': 'tac', 'value': 100.0, 'groups': 'snare', 'type': 'child',
         'order': 1.0},
        {'variation': inf, 'label': 'Tom', 'value': 1.0, 'groups': 'tom', 'type': 'parent',
         'order': nan},
        {'variation': inf, 'label': 'bom', 'value': 1.0, 'groups': 'tom', 'type': 'child',
         'order': 1.0},
        {'variation': nan, 'label': 'Trimester 2', 'value': 111.0, 'groups': nan, 'type': nan,
         'order': nan}
    ]
    df = pd.DataFrame(data)
    df = waterfall(df, **kwargs)
    wa = [{k: v for k, v in zip(df.columns, row)} for row in df.values]
    assert wa[0].keys() == expected[0].keys()
    for i in range(len(expected)):
        testing.assert_equal(wa[i], expected[i])


def test_waterfall_null():
    kwargs = {
        'upperGroup': {'id': 'category_id', 'label': 'category_name'},
        'insideGroup': {'id': 'product_id', 'groupsOrder': 'ord'},
        'date': 'date',
        'value': 'played',
        'start': {'label': 'Trimestre 1', 'id': 't1'},
        'end': {'label': 'Trimester 2', 'id': 't2'},
    }

    data = None
    df = pd.DataFrame(data)
    df = waterfall(df, **kwargs)
    wa = [{k: v for k, v in zip(df.columns, row)} for row in df.values]
    assert wa == []


def test_waterfall_not_implemented(sample_data):
    """ It should raise Error for not implemented features """
    kwargs = {
        'upperGroup': {'id': 'category_id', 'label': 'category_name'},
        'insideGroup': {'id': 'product_id', 'groupsOrder': 'ord'},
        'date': 'date',
        'value': 'played',
        'start': {'label': 'Trimestre 1', 'id': 't1'},
        'end': {'label': 'Trimester 2', 'id': 't2'},
        'breakdown': ['id']
    }
    df = pd.DataFrame(sample_data)
    with pytest.raises(NotImplementedError) as exc_info:
        waterfall(df, **kwargs)
    assert str(exc_info.value) == 'We will add breakdown support ' \
                                  'on your request, please contact the devs'

    kwargs = {
        'upperGroup': {'id': 'category_id', 'label': 'category_name'},
        'date': 'date',
        'value': 'played',
        'start': {'label': 'Trimestre 1', 'id': 't1'},
        'end': {'label': 'Trimester 2', 'id': 't2'},
    }
    df = pd.DataFrame(sample_data)
    with pytest.raises(NotImplementedError) as exc_info:
        waterfall(df, **kwargs)
    assert str(exc_info.value) == 'We will add support for upperGroup only ' \
                                  'on you request, please contact the devs'



from typing import List


def check_params_columns_duplicate(cols_name: List[str]) -> bool:
    params = [column for column in cols_name if column is not None]
    if len(set(params)) != len(params):
        duplicates = set([x for x in params if params.count(x) > 1])
        raise ParamsValueError(
            f'Duplicate declaration of column(s) {duplicates} in the parameters')
    else:
        return True


class ParamsValueError(Exception):
    """
    Exception raised when some parameters value are wrong
    """



import numpy as np
import pandas as pd


def replace(df, column, dst_column=None, **kwargs):
    """
    Replace values of a column (uses pandas.Series.replace)
    Args:
        df (pd.DataFrame): DataFrame to transform
        column (str): name of the column containing values to replace
        dst_column (str): optional, name of the column which will contain replaced
                          values (same as "column" by default)

        Other parameters are directly forwarded to pandas.Series.replace.
    """
    dst_column = dst_column or column
    df.loc[:, dst_column] = df[column].replace(**kwargs)
    return df


def rename(df, values=None, columns=None, locale=None):
    """
    Replaces data values and column names according to locale
    Args:
        df (pd.DataFrame): DataFrame to transform
        values (dict):
            - key (str): term to be replaced
            - value (dict):
                - key: locale
                - value: term's translation
        columns (dict):
            - key (str): columns name to be replaced
            - value (dict):
                - key: locale
                - value: column name's translation
        locale (str): locale
    """
    if values:
        to_replace = list(values.keys())
        value = [values[term][locale] for term in values]
        df = df.replace(to_replace=to_replace, value=value)
    if columns:
        _keys = list(columns.keys())
        _values = [column[locale] for column in columns.values()]
        columns = dict(list(zip(_keys, _values)))
        df = df.rename(columns=columns)
    return df


def melt(df, id, value, dropna=False):
    """
    This function is useful to massage a DataFrame into a format where one or more columns
    are identifier variables (id), while all other columns,
    considered measured variables (value), are “unpivoted” to the row axis,
    leaving just two non-identifier columns, ‘variable’ and ‘value’.
    Args:
        df (pd.DataFrame): DataFrame to transform
        id (list): Column(s) to use as identifier variables
        value (list): Column(s) to unpivot.
        dropna (bool): dropna in added 'value' column
    """
    try:
        df = df[(id + value)]
        df = pd.melt(df, id_vars=id, value_vars=value)
        if dropna:
            df = df.dropna(subset=['value'])
    except KeyError as e:
        raise Exception(f'Invalid configuration for melt, missing key: {e}')
    else:
        return df


def top(df, value, limit, order='asc', group=None):
    """
    Awesome method that achieves what NO query in any language can do: (DRUM ROLL)
    Get the top or flop N results based on a column value for each specified group columns
    Args:
        - group: String or array of strings for the columns,
                 on which you want to perform the group operation
        - value: String for the column name on which you will rank the results
        - order: String 'asc' or 'desc' to sort by ascending ou descending order
        - limit: Number to specify the N results you want to retrieve.
                 Use a positive number x to retrieve the first x results.
                 Use a negative number -x to retrieve the last x results.
    """
    ascending = order != 'desc'
    filter_func = 'nlargest' if (limit > 0) ^ ascending else 'nsmallest'

    def _top(df):
        return getattr(df, filter_func)(abs(limit), value).sort_values(by=value,
                                                                       ascending=ascending)

    try:
        if group is None:
            df = _top(df)
        else:
            df = df.groupby(group).apply(_top)
    except KeyError as e:
        raise Exception(f'Invalid configuration for top, missing key: {e}')
    else:
        return df


def pivot(df, index, column, value):
    """
    Pivot a dataframe. Reverse operation of melting. Useful for configuring evolution
    See pandas' pivot_table documentation for more details
    Args:
        - index (list): indexes argument of pd.pivot_table
        - column (str): column name to pivot on
        - value (str): column name containing the value to fill the pivoted df
    """
    try:
        if df.dtypes[value].type == np.object_:
            df = pd.pivot_table(df, index=index,
                                columns=column,
                                values=value,
                                aggfunc=lambda x: ' '.join(x))
        else:
            df = pd.pivot_table(df, index=index,
                                columns=column,
                                values=value)
        df = df.reset_index()
    except KeyError as e:
        raise Exception(f'Invalid configuration for pivot, missing key: {e}')
    else:
        return df


def pivot_by_group(df, variable, value, new_columns, groups, id_cols=None):
    if id_cols is None:
        index = [variable]
    else:
        index = [variable] + id_cols

    param = pd.DataFrame(groups, index=new_columns)
    temporary_colum = 'tmp'

    df[temporary_colum] = df[variable]
    for column in param.columns:
        df.loc[df[variable].isin(param[column]), variable] = column

    param = param.T
    for column in param.columns:
        df.loc[
            df[temporary_colum].isin(param[column]), temporary_colum] = column

    df = pivot(df, index, temporary_colum, value)
    return df


def argmax(df, column):
    df = df[df[column] == df[column].max()]
    return df


def fillna(df, column, value):
    """
    Can fill NaN values from a column
    Args:
        df
        column
        value
    """
    if column in df.columns:
        df[column] = df[column].fillna(value)
    else:
        df[column] = value
    return df


def query_df(df, query):
    """
    Slice the data according to the provided query
    Basic usage like the one in the data query but in the postprocess
    Useful if you want to perform some slicing after a melt or pivot for example.
    Wired on the DataFrame.query method, see doc
    http://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.query.html#pandas.DataFrame.query
    Args:
        Query String
    """
    try:
        df = df.query(query)
    except Exception as e:
        raise Exception(f'Invalid query: {e}')
    else:
        return df


def add(df, new_column, column_1, column_2):
    """
    Basic mathematical operation to add column_1 and column_2 values.
    Will create a new column named `new_column`
    """
    try:
        df[new_column] = df[column_1] + df[column_2]
    except KeyError as e:
        raise Exception(f'Invalid config for sum: {e}')
    else:
        return df


def subtract(df, new_column, column_1, column_2):
    """
    Basic mathematical operation to substract column_2 to column_1 values.
    Will create a new column named `new_column`
    """
    try:
        df[new_column] = df[column_1] - df[column_2]
    except KeyError as e:
        raise Exception(f'Invalid config for subtract: {e}')
    else:
        return df


def multiply(df, new_column, column_1, column_2):
    """
    Basic mathematical operation to multiply column_1 and column_2 values.
    Will create a new column named `new_column`
    """
    try:
        df[new_column] = df[column_1] * df[column_2]
    except KeyError as e:
        raise Exception(f'Invalid config for multiply: {e}')
    else:
        return df


def divide(df, new_column, column_1, column_2):
    """
    Basic mathematical operation to divide column_2 to column_1 values.
    Will create a new column named `new_column`
    """
    try:
        df[new_column] = df[column_1] / df[column_2]
    except KeyError as e:
        raise Exception(f'Invalid config for divide: {e}')
    else:
        return df


def cumsum(df, new_column, column, index, date_column, date_format):
    """
    Creates a new column, which is the cumsum of the column
    :param df: the dataframe
    :param new_column: name of the new column
    :param column: name on which the cumulative sum is performed
    :param index: array of column names to keep as indices
    :param date_column: column name that represent the date
    :param date_format: format of the date
    :return:
    """
    date_temp = 'date_temp_'
    if isinstance(index, str):
        index = [index]
    try:
        df[date_temp] = pd.to_datetime(df[date_column], format=date_format)
        df2 = (df.groupby(index)
               .apply(lambda x: x.set_index(date_temp)
                      .sort_index()[[column]]
                      .cumsum())
               .reset_index()
               .rename(columns={column: new_column})
               )
        if new_column == column:
            df.drop(column, axis=1, inplace=True)
        df = df.merge(df2, on=[*index, date_temp]).drop(date_temp, axis=1)
    except KeyError as e:
        raise Exception(f'Invalid config for cumsum: {e}')
    else:
        return df


def percentage(df, new_column, column, group_cols=None):
    """
    Add a column to the dataframe according to the groupby logic on group_cols
    :param df: Dataframe
    :param new_column: name of the new column
    :param column: name of the desired column you need percentage on
    :param group_cols: (str | list of str) or None
    :return: df + the percentage column
    """
    if group_cols is None:
        df[new_column] = 100. * df[column] / sum(df[column])
    else:
        df[new_column] = (df.groupby(group_cols)
                          .apply(lambda x: 100. * x[column] / sum(x[column]))
                          .sort_index(level=len(group_cols))
                          .reset_index()[column]
                          )
    return df


def waterfall(df, date, value, start, end, upperGroup,  # noqa:C901
              insideGroup=None, breakdown=None):
    """
    Return a line for each bars of a waterfall chart, totals, groups, subgroups.
    Compute the variation and variation rate for each line.

    Args:
        upperGroup (dict)
            - id: name of the column that contains upperGroups unique IDs
            - label: not required, text displayed under each upperGroups bars,
                     using ID when it's absent
            - groupsOrder: not required, order of upperGroups
        insideGroup (dict)
            - id: name of the column that contains insideGroups unique IDs
            - label: not required, text displayed under each insideGroups bars,
                     using ID when it's absent
            - groupsOrder: not required, order of insideGroups
        date (str): name of the column that id the period of each lines
        value (str): name of the column that contains the vaue for each lines
        start (dict):
            - label: text displayed under the first master column
            - date: value in the date col that id lines for the first period
        end (dict):
            - label: text displayed under the last master column
            - date: value in the date col that id lines for the second period
        breakdown (list) : (not implemented)
        # fillValues (bool): (case when false is not implemented)

    TODO:
        * idea to implement breakdown, call waterfall() on data filtered by filter.
        * crash gracefully if data or config is not coherent
        * il y aura des cols en trop: les gérer
    """

    def rename_or_create(group, group_name, name, target, default):
        """
        Rename or create standard columns for optional options
        """
        col_name = name.format(group_name)
        if not group.get(target):
            df[col_name] = default
        else:
            df.rename(columns={group[target]: col_name}, inplace=True)

    def first(x):
        return x.iloc[0]

    def compute_totals(totals):
        """
        Compute two dataframes with value for start and end
        Args:
            totals(dataframe):

        Returns: Dataframe, Dataframe

        """
        result = {}
        time_dict = {'start': start, 'end': end}
        for time_name, time in time_dict.items():
            if not totals[totals['date'] == time['id']].empty:
                value = totals.loc[
                    totals['date'] == time['id'], 'value'
                ].values[0]
            else:
                value = 0
            result[time_name] = pd.DataFrame([{
                'value': value,
                'label': time['label']
            }])
        return result['start'], result['end']

    def merge_df(df):
        """
        Compute diff value between start and end
        Args:
            df(dataframe):

        Returns: Dataframe

        """
        start_df, end_df = df[df['date'] == start_id], df.loc[df['date'] == end_id]
        df = start_df.merge(end_df,
                            on=['insideGroup', 'upperGroup',
                                'upperGroup_order',
                                'insideGroup_order', 'upperGroup_label',
                                'insideGroup_label'],
                            how='outer',
                            suffixes=('_start', '_end'), )

        # necessary before calculating variation
        df[['value_start', 'value_end']] = df[['value_start', 'value_end']].fillna(0)
        df['value'] = df['value_end'] - df['value_start']
        df.drop(['date_start', 'date_end', 'value_end'], axis=1, inplace=True)
        df.rename(columns={'upperGroup': 'groups'}, inplace=True)
        return df

    def compute_inside_group(df):
        """
        Compute inside Group
        Args:
            df(dataframe):

        Returns: Dataframe

        """
        inside_group = df.copy()
        inside_group['type'] = 'child'
        inside_group['variation'] = inside_group['value'] / inside_group['value_start']
        inside_group.drop(['upperGroup_label', 'insideGroup', 'value_start'],
                          axis=1, inplace=True)
        inside_group.rename(columns={'insideGroup_label': 'label'}, inplace=True)
        return inside_group

    def compute_upper_group(df):
        """
        Compute upperGroup
        Args:
            df (Dataframe):

        Returns: Dataframe

        """
        upper_group = df.groupby(['groups']).agg({
            'value': sum,
            'value_start': sum,
            'upperGroup_label': first,
            'upperGroup_order': first
        }).reset_index()
        upper_group['type'] = 'parent'
        upper_group['variation'] = upper_group['value'] / upper_group['value_start']
        upper_group.drop(['value_start'], axis=1, inplace=True)
        upper_group.rename(columns={'upperGroup_label': 'label'}, inplace=True)
        return upper_group

    if len(df) == 0:
        return df

    start_id, end_id = start['id'], end['id']

    if breakdown is not None:
        raise NotImplementedError(
            'We will add breakdown support on your request, please contact the devs')
    if insideGroup is None:
        raise NotImplementedError(
            'We will add support for upperGroup only on you request, please contact the devs')

    # prepare the dataframe with standard column names

    df = df.rename(columns={
        upperGroup['id']: 'upperGroup',
        insideGroup['id']: 'insideGroup',
        date: 'date',
        value: 'value'
    })
    groups_dict = {'upperGroup': upperGroup, 'insideGroup': insideGroup}
    for g_name, g in groups_dict.items():
        rename_or_create(g, g_name, '{}_label', 'label', df[g_name])
        rename_or_create(g, g_name, '{}_order', 'groupsOrder', pd.np.nan)

    agg_conf = {'value': sum}
    agg_conf.update({'{}_label'.format(col): first for col in groups_dict.keys()})
    agg_conf.update({'{}_order'.format(col): first for col in groups_dict.keys()})
    df = df.groupby([*groups_dict.keys(), 'date']).agg(agg_conf).reset_index()
    start, end = compute_totals(df.groupby('date').agg({'value': sum}).reset_index())

    df = merge_df(df)

    groups = pd.concat([
        compute_upper_group(df),
        compute_inside_group(df)
    ]).sort_values(
        by=['upperGroup_order', 'groups', 'type', 'insideGroup_order', 'label'],
        ascending=[True, True, False, True, True])
    ret = pd.concat([start, groups, end])

    # We don't need specific sorting orders by groups now
    ret.loc[ret['type'] == 'child', 'order'] \
        = ret.loc[ret['type'] == 'child', 'insideGroup_order']
    ret.loc[ret['type'] == 'parent', 'order'] \
        = ret.loc[ret['type'] == 'parent', 'upperGroup_order']
    ret.drop(['insideGroup_order', 'upperGroup_order'], axis=1, inplace=True)

    return ret

query = query_df
